# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# wdic-fpw.pl -- FreePWING script for WDIC
# $Id: wdic-fpw.pl 1.5 2004/12/11 21:16:01 satomii Exp $
#
# DO NOT CHANGE THE KANJI CODE OF THIS FILE.
#
# Copyright (C) 2000-2004, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use strict;
use Getopt::Long;
use File::DosGlob 'glob';
use IO::File;
use FreePWING::FPWUtils::FPWParser;
use Jcode;
use Wdic;

require 'wdic-reform.pl';
require 'wdic-menu.pl';
require 'wdic-body.pl';

use vars qw(
	$VERSION
	$opt_bullet $opt_gaiji $opt_dirinfo $opt_cgraphs $opt_sounds $opt_plugin
	$wdic $wdic_dir $newline_on_death %cgraphs %plugins %sounds
	$fpwtext $fpwheading $fpwword2 $fpwkeyword $fpwmenu $fpwcopyright
);

$VERSION = '$Revision: 1.5 $';

$wdic_dir = '.';
$opt_bullet = '';
$opt_gaiji = 1;
$opt_dirinfo = 0;
$opt_cgraphs = '';
$opt_sounds = '';
$opt_plugin = '';

$newline_on_death = 0;

#
# signal hooks.
#
$SIG{__DIE__} = sub {
	my $err = $_[0];

	$err = "Error: " . Wdic::outstr($err);
	if ($newline_on_death) {
		$err = "\n$err";
		$newline_on_death = 0;
	}
	if ($Wdic::LOGFILEHANDLE) {
		print $Wdic::LOGFILEHANDLE "$err\n";
	}
	die $err;
};

$SIG{__WARN__} = sub {
	my $err = $_[0];

	$err = "Warning: " . Wdic::outstr($err);
	if ($newline_on_death) {
		$err = "\n$err";
		$newline_on_death = 0;
	}
	if ($Wdic::LOGFILEHANDLE) {
		print $Wdic::LOGFILEHANDLE $err;
	}
	warn $err;
};

##------------------------------------------------------------------------
## ENTRY POINT

my $word_count;

&initialize();

$wdic = new Wdic;
$word_count = $wdic->init_wdic($wdic_dir);
&reform_class_data();

initialize_fpwparser(
	'text' => \$fpwtext,
	'heading' => \$fpwheading,
	'word2' => \$fpwword2,
	'keyword' => \$fpwkeyword,
	'menu' => \$fpwmenu,
	'copyright' => \$fpwcopyright
);

&make_menu();
&make_copyright();
&make_body($word_count);
&make_plugin_list();

finalize_fpwparser(
	'text' => \$fpwtext,
	'heading' => \$fpwheading,
	'word2' => \$fpwword2,
	'keyword' => \$fpwkeyword,
	'menu' => \$fpwmenu,
	'copyright' => \$fpwcopyright
);

&uninitialize();
exit;

##------------------------------------------------------------------------
## INITIALIZATION

sub initialize()
{
	my @argv = @ARGV;
	my ($charset, $log, $usage);
	local $_;

	if (!GetOptions('help|h|?' => \$usage,
			'charset|c:s' => \$charset,
			'dirinfo!' => \$opt_dirinfo,
			'bullet|b:s' => \$opt_bullet,
			'gaiji|g:i' => \$opt_gaiji,
			'cgraphs:s' => \$opt_cgraphs,
			'sounds:s' => \$opt_sounds,
			'plugin:s' => \$opt_plugin,
			'log:s' => \$log)
			|| 1 < scalar(@ARGV)) {
		&usage(1);
	}
	&usage(0) if ($usage);

	#
	# if --log=FILE option is used, create an log file for debug purpose.
	# the existing file is overwritten without confirmation.
	#
	if ($log) {
		$Wdic::LOGFILEHANDLE = new IO::File($log, 'w')
			|| die "Unable to create log file: $log ($!)\n";
		$Wdic::LOGFILEHANDLE->print(
				"# Started by ", &user(), " on ", &now(), "\n",
				"# Command line: $0 @argv\n\n");
	}

	#
	# check the input directories. the current directory must contain
	# readme.txt, and the WDIC directory must contain WDICALL.MAN (which
	# was named as WDIC.MAN in some older versions of WDIC, but this old
	# name is no longer supported by wdic-fpw1.3c or later).
	#
	$wdic_dir = shift(@ARGV) if (@ARGV);
	$wdic_dir =~ s/\\/\//g;
	$wdic_dir =~ s/\/$//;
	-d $wdic_dir || die "Invalid input directory: $wdic_dir ($!).\n";

	if (!glob("$wdic_dir/WDICALL.MAN") && !(-f "$wdic_dir/WDICALL.MAN")) {
		die "File not found: $wdic_dir/WDICALL.MAN ($!)\n";
	}
	if (!glob("./readme.txt") && !(-f "./readme.txt")) {
		die "File not found: ./readme.txt ($!)\n";
	}

	if ($opt_plugin) {
		$opt_plugin =~ s/\\/\//g;
		$opt_plugin =~ s/\/$//;
		-d $opt_plugin || die "Invalid plugin directory: $opt_plugin ($!).\n";
	}

	#
	# validate the charset option, or use the environment value if not
	# specified at the command line.
	#
	if (!$charset) {
		if ($ENV{LANG} && $ENV{LANG} =~ /^ja_JP\.([a-zA-Z]+)$/) {
			$charset = $1;
		} else {
			$charset = '';
		}
	}
	$charset =~ tr/A-Z/a-z/;
	if ('ujis' eq $charset || 'eucjp' eq $charset) {
		$charset = 'euc';
	} elsif (!$charset) {
		if ((!$ENV{OS}
				|| $ENV{OS} !~ /^Windows/)
				&& (!$ENV{COMSPEC}
				|| $ENV{COMSPEC} !~ /(COMMAND\.COM)|(CMD\.EXE)$/i)) {
			$charset = 'euc';
		} else {
			$charset = 'sjis';
		}
	} elsif ('sjis' ne $charset
			&& 'euc' ne $charset
			&& 'jis' ne $charset
			&& 'iso_2022_jp' ne $charset) {
		die "Unknown character set: $charset\n";
	}
	&log("Detected display charset: $charset\n");
	$Wdic::CHARSET = $charset;

	#
	# ensure that other option values are valid.
	#
	$opt_bullet = '' if (!defined($opt_bullet) || '' eq $opt_bullet);
	$opt_bullet = Jcode::convert($opt_bullet, 'euc');

	#
	# pre-load cgraphs.txt and sounds.txt if necessary. keys of %cgraphs
	# and %sounds indicate the binary IDs.
	#
	&load_binary_list($opt_cgraphs, '.bmp', \%cgraphs) if ($opt_cgraphs);
	&load_binary_list($opt_sounds, '.wav', \%sounds) if ($opt_sounds);
}

sub uninitialize()
{
	$wdic->uninit_wdic();
}

#
# (hash) load_binary_list(char *file, char *ext)
#
sub load_binary_list($$\%)
{
	my ($file, $ext, $href) = @_;
	my @pair;

	open(LIST, "<$file") || die "Unable to open file: $file ($!)\n";
	&log("Reading $file...");
	$newline_on_death = 1;
	$ext = qr($ext);

	while (<LIST>) {
		s/[ \t\r\n]+$//;
		@pair = split(/[ \t]+/, $_);
		if (1 == $#pair) {
			$_ = $pair[1];
			s/^(.*[\/\\])?([^\/\\]+)$ext$/$2/;
			$$href{$pair[0]}{name} = $_;
			$$href{$pair[0]}{list} = [];
		}
	}

	close(LIST);
	&log(" done.\n");
	$newline_on_death = 0;
}

#
# void reform_class_data(void)
# modifies and/or adds necessary fields in Wdic object.
#
sub reform_class_data()
{
	my (%data, @keys);
	local $_;

	%data = %{ $wdic->{wdic_file} };
	@keys = keys(%data);
	foreach (@keys) {
		$data{$_}{NAME} = &reform_sjis($data{$_}{NAME}, 0);
		$data{$_}{NAME} =~ s/ //g;
	}
	$wdic->{wdic_file} = { %data };

	%data = %{ $wdic->{wdic_group} };
	@keys = keys(%data);
	foreach (@keys) {
		$data{$_}{NAME} = &reform_sjis($data{$_}{NAME}, 0);
		$data{$_}{NAME} =~ s/ //g;
	}
	$wdic->{wdic_group} = { %data };

	%data = %{ $wdic->{wdic_dirs} };
	@keys = keys(%data);

	foreach (@keys) {
		$data{$_}{ID} || die "Directory ID not defined: $_";
		$data{$_}{NAME} = &reform_sjis($data{$_}{NAME}, 0);

		# even if this directory is directed to another directory, the target
		# may not be defined in DIR.LST.
		my $ref = $data{$_}{REF};
		if ($ref) {
			if ($data{$ref}) {
				$data{$_}{TAG} = sprintf('DIR%04x', $data{$ref}{ID});
				$data{$_}{NAME} .= '@';
			} else {
				warn "Invalid directory link: $_ => $ref\n";
				$data{$_}{REF} = undef;
				$data{$_}{TAG} = sprintf('DIR%04x', $data{$_}{ID});
			}
		} else {
			$data{$_}{TAG} = sprintf('DIR%04x', $data{$_}{ID});
		}

		$data{$_}{NAME_EUC} = Jcode::convert($data{$_}{NAME}, 'euc', 'sjis');
		$data{$_}{DIR} = $_;
	}

	$wdic->{wdic_dirs} = { %data };
}

##------------------------------------------------------------------------
## OUTPUT FUNCTIONS

#
# bool write_sjis($obj, char *in)
# writes SJIS text using the specified class object.
#
sub write_sjis($$)
{
	my ($fpwobj, $in) = @_;
	my ($newline, $out, $c);

	$newline = ($in =~ s/([^\r\n]*)[\r\n]+$/$1/);

	if ($opt_gaiji < 2) {
		# do not use user characters.
		$out = Jcode::convert($in, 'euc', 'sjis');
		$out =~ s/\\/\xa1\xef/g;	#  in EUC.
		$fpwobj->add_text($out)
			|| die &fpwerr("Unable to write text: $out", $fpwobj);
		if ('' eq $in || $newline) {
			$fpwobj->add_newline()
				|| die &fpwerr("Unable to add newline", $fpwobj);
		}
		return 1;
	}

	$out = '';

	while ('' ne $in) {
		$c = substr($in, 0, 1);
		$in = substr($in, 1);

		if ($c =~ /[\x81-\x9f\xe0-\xfc]/) {
			# SJIS double byte characters.
			$out .= $c;
			$out .= substr($in, 0, 1);
			$in = substr($in, 1);

		} elsif ($c =~ /[\x5c\xa1-\xdf]/) {
			# '\' or JIS X 0201.
			my $gaiji = sprintf("sj_%02x", unpack('C', $c));
			if ('' ne $out) {
				$fpwobj->add_text(Jcode::convert($out, 'euc', 'sjis'))
					|| die &fpwerr("Unable to write text: $out", $fpwobj);
				$out = '';
			}
			$fpwobj->add_half_user_character($gaiji)
				|| die &fpwerr("Unable to add gaiji: $gaiji", $fpwobj);
		} else {
			# ASCII characters.
			$out .= $c;
		}
	}

	if ('' ne $out) {
		$fpwobj->add_text(Jcode::convert($out, 'euc', 'sjis'))
			|| die &fpwerr("Unable to write text: $out", $fpwobj);
	}
	if ('' eq $_[1] || $newline) {
		$fpwobj->add_newline()
			|| die &fpwerr("Unable to add newline", $fpwobj);
	}
	1;
}

#
# bool write_newline($fpwobj, int count)
# writes newline using the specified object.
#
sub write_newline($;$)
{
	my ($fpwobj, $count) = @_;
	my $i;

	$count = 1 unless defined($count);
	for ($i = 0; $i < $count; $i++) {
		$fpwobj->add_newline()
			|| die &fpwerr("Unable to add newline", $fpwobj);
	}
	$i;
}

#
# bool write_gaiji($fpwobj, char **gaiji)
# inserts half-width user characters.
#
sub write_gaiji($@)
{
	my ($fpwobj, @gaiji) = @_;
	local $_;

	foreach (@gaiji) {
		$fpwobj->add_half_user_character($_)
			|| die &fpwerr("Unable to add gaiji: $_", $fpwobj);
	}
	1;
}

#
# bool write_link($fpwobj, char *label, char *tag)
# adds a cross-reference to another entry using the specified class object.
# the label is assumed to be SJIS string.
#
sub write_link($$$)
{
	my ($fpwobj, $label, $tag) = @_;
	my $newline = ($label =~ s/([^\r\n])[\r\n]+$/$1/);

	$tag || die "Invalid reference tag: (empty) for \"$label\"";
	if (!$fpwobj->add_reference_start()
			|| !$fpwobj->add_text(Jcode::convert($label, 'euc', 'sjis'))
			|| !$fpwobj->add_reference_end($tag)) {
		die &fpwerr("Unable to add reference: $label", $fpwobj);
	}
	&write_newline($fpwobj, 1) if ($newline);
	1;
}

#
# char **write_word_link($fpwobj,
#      char *bullet, char *word, char *group, char *file)
# searches another word in the current source files and links to it if
# one is found. returns the target group and file (dictionary) if found.
#
sub write_word_link($$$;$$)
{
	my ($fpwobj, $bullet, $word, $group, $file) = @_;
	my $newline = ($word =~ s/([^\r\n])[\r\n]+$/$1/);
	my @list = $wdic->locate_word($word, $group, $file);

	if (4 == scalar(@list)) {
		# @list = (index, group, file, offset)
		&write_sjis($fpwobj, $bullet) if ('' ne $bullet);
		&write_link($fpwobj, $word, sprintf('%s%08x', $list[2], $list[3]));
		@list = ($list[1], $list[2]);
	} else {
		&write_sjis($fpwobj, "$bullet$word");
		@list = ();
	}

	&write_newline($fpwobj, 1) if ($newline);
	@list;
}

#
# void write_dir_link($fpwobj, char **list)
# writes a list of directories with cross-references as "DIR1 > DIR2 ...".
#
sub write_dir_link($$@)
{
	my ($fpwobj, $mode, @dirs) = @_;
	my (%info, @list);
	local $_;

	foreach (@dirs) {
		my $path = '';
		my $this;

		@list = split(/\/+/, $_);
		while (@list && !$list[0]) { shift(@list) };
		while (@list && !$list[$#list]) { pop(@list) };
		$this = pop(@list) if ($fpwobj == $fpwmenu);
		&write_link($fpwobj, '', 'DIR');

		while (@list) {
			$fpwobj->add_text($path ? ' > ' : ': ');
			$path .= "/$list[0]";
			%info = $wdic->get_dir_info($path);
			if (%info) {
				$info{TAG} || die "Directory tag not defined: $path";
				if (!$fpwobj->add_reference_start()
						|| !$fpwobj->add_text($info{NAME_EUC})
						|| !$fpwobj->add_reference_end($info{TAG})) {
					die &fpwerr("Unable to add reference: " .
							"$info{NAME_EUC} ($info{TAG})",
							$fpwobj);
				}
			} else {
				warn "Unknown directory: $path\n";
				$fpwobj->add_text($list[0])
					|| die &fpwerr("Unable to write text: $dirs[0]", $fpwobj);
			}
			shift(@list);
		}

		if ($this) {
			%info = $wdic->get_dir_info($_);
			if ($info{TAG}) {
				$this = $info{NAME_EUC};
			} else {
				warn "Unknown directory: $_\n";
			}
			$fpwobj->add_text(" > $this")
				|| die &fpwerr("Unable to write text:  > $this", $fpwobj);
		}
		&write_newline($fpwobj, 1);
	}
}

##------------------------------------------------------------------------
## MISC FUNCTIONS

sub new_file_handle($$)
{
	my ($file, $mode) = @_;
	my $handle = new IO::File($file, $mode)
		|| die "Unable to open file: $file ($!)";
	binmode $handle;
	$handle;
}

sub log(@)
{
	Wdic::log(@_);
}

sub user()
{
	if (defined($ENV{'USER'})) {
		$ENV{'USER'};
	} elsif (defined($ENV{'USERNAME'})) {
		$ENV{'USERNAME'};
	} else {
		'(unknown user)';
	}
}

sub now()
{
	my ($sec, $min, $hour, $day, $mon, $year) = localtime(time);
	sprintf("%04d/%02d/%02d %02d:%02d:%02d",
			$year + 1900, $mon + 1, $day, $hour, $min, $sec);
}

sub fpwerr($$)
{
	my ($str, $fpwobj) = @_;

	if ($fpwobj) {
		my $err = $fpwobj->error_message();
		$str .= " ($err)" if ($err);
	}
	return "$str,";
}

sub usage($)
{
	eval { require Pod::Usage };
	if ($@) {
		print<<__EOF__;
Usage: perl $0 [-c CHARSET] [-b BULLET] [-g LEVEL]
	    [--cgraphs=FILE] [--plugin=DIR] [--dirinfo] [INPUT_DIR]
If inputdir is omitted, the current directory (.) is used.
For more information, use pod2man(1) or any similar utility.

__EOF__
	} else {
		Pod::Usage::pod2usage(-verbose => 1);
	}
	exit($_[0]);
}

__END__

##-------------------------------------------------------------------------
## DOCUMENTATION

=pod

=head1 NAME

wdic-fpw.pl - a FreePWING script for WDIC

=head1 SYNOPSIS

perl wdic-fpw.pl [-c CHARSET] [-b BULLET] [-g GAIJI] [--cgraphs=FILE]
[--plugin=DIR} [INPUT_DIR]

=head1 DESCRIPTION

This script generates a JIS X 4081 book from 'WDIC' (possibly
"WIRED & WIRELESS DICTIONARY") using FreePWING. The generated book can be
viewed with many EPWING viewers.

=head1 OPTIONS

=over 8

=item B<-h> | B<-?> | B<--help>

Displays this usage.

=item B<-c> I<CHARSET> | B<--charset>=I<CHARSET>

Specifies the display charset, which is used when printing progress. The
default is I<sjis> for Windows and I<euc> for others. This option would never
affect the book-producing routines.

=item B<-b> I<BULLET> | B<--bullet>=I<BULLET>

Specifies a character or string to be used as a 'bullet' for each description.
The default is a full-width white diamond.

=item B<-g> I<LEVEL> | B<--gaiji>=I<LEVEL>

Use half-shape user-defined characters (gaiji). Valid levels are:
0) Do not use any user characters;
1) Use user characters for some phonetic symbols;
2) Use user characters for some phonetic symbols, signle-byte yen marks ('\')
and so-colled 'hankaku katakana's.

=item B<--cgraphs>=I<FILE>

Specifies the path to the cgraph.txt file.
If this option is used, graphic plugins listed in FILE are embedded into the
resulting book.
FILE must be the one generated by Makefile that comes with this script.

=item B<--sounds>=I<FILE>

Specifies the path to the sounds.txt file.
If this option is used, sound plugins listed in FILE are embedded into the
resulting book.
FILE must be the one generated by Makefile that comes with this script.

=item B<--plugin>=I<DIR>

Specifies the directory where text and sound plugin files are located.
Some text-based plugins in this directory may be embedded to the book.

=item B<--dirinfo>

Writes 'directory information' into the body.
Some entries in WDIC may have the information that indicates what directory
(category) it belongs to.
If this option is used, such information (if available) is added before the
description lines.

=back

=head1 B<ENVIRONMENT>

=over 8

=item B<LANG>

Referred when B<--charset> option is not given.

=item B<OS> or B<COMSPEC> or B<ComSpec>

Referred when B<--charset> option is not given and B<LANG> is not defined, to
distinguish the Windows systems.

=item B<USER> or B<USERNAME>

Used to record the build information into the copyright page. If these
variables were not defined, the builder's name would default to
'(unknown user)'.

=back

=head1 B<FILES>

=item B<FreePWING>

A set of Perl modules to generate a JIS X 4081 book. Visit
E<lt>http://www.sra.co.jp/people/m-kasahr/freepwing/E<gt> for more information.

=item B<Jcode.pm>

A Perl module for kanji-code conversion. Visit
E<lt>http://openlab.ring.gr.jp/Jcode/E<gt> for more information.

=item B<Wdic.pm>

A Perl module to read entries from WDIC files.

=item B<wdic-menu.pl>

A Perl script used from wdic-fpw main script, to construct the menu and
copyright entries.

=item B<wdic-body.pl>

A Perl script used from wdic-fpw main script, to write all word entries into
the book.

And, of course, the original WDIC files are also required to run wdic-fpw.

=head1 VERSION

$Id: wdic-fpw.pl 1.5 2004/12/11 21:16:01 satomii Exp $

=head1 AUTHOR

rei@wdic.org

=head1 SEE ALSO

=item B<WIRED & WIRELESS DICTIONARY>

http://www.wdic.org/

=cut
